#!/usr/bin/perl -w

use strict;

my %opt_is_race_array;  # value is pretty-printed name
my %opt_value;
my %opt_modified;
my %opt_pretty;
my $template_gripe = 0;
my @output_template;    # each entry is either a config option name, or a "+" followed by some text to output

if (!@ARGV) {
    print "usage: $0 [src=pconfig.src] [def=config.hi4] [N,N=file.txt]

def=config.hi4: name of file containing option definitions (config.hi4)

src=pconfig.src: output template and option values (pconfig.src)

N,N=file.txt: read race N's definition from file.txt\n";
    exit 1;
}

foreach (@ARGV) {
    if (m{^src=(.*)}) {
        my $reading = 1;
        my $read_as_template = (@output_template == 0);
        open FILE, "< $1" or die "$1: $!";
        while (<FILE>) {
            chomp;
            if (/^\s*\#/ || /^\s*$/) {
                push @output_template, "+$_"
                  if $read_as_template;
            } elsif (/^\s*%\s*phost/i) {
                $reading = 1;
                push @output_template, "+$_"
                  if $read_as_template;
            } elsif (/^\s*%/) {
                $reading = 0;
                push @output_template, "+$_"
                  if $read_as_template;
            } elsif ($reading && /^\s*(\S+)\s*=\s*([^\#]*)/) {
                $opt_value{lc($1)} = $2;
                $opt_pretty{lc($1)} = $1;
                push @output_template, lc($1)
                  if $read_as_template;
            } elsif ($reading) {
                die "syntax error: `$_'";
            } else {
                push @output_template, "+$_"
                  if $read_as_template;
            }
        }
        close FILE;
    } elsif (m{^def=(.*)}) {
        open FILE, "< $1" or die "$1: $!";
        while (<FILE>) {
            if (/CFDefine4\(([a-z0-9]+).*"(.*)"\)/i) {
                my ($name, $value) = ($1, $2);
                $opt_value{lc($name)} = $value;
                $opt_pretty{lc($name)} = $name;
                if (/RACEARRAY/) {
                    $opt_is_race_array{lc($name)} = 1;
                }
            }
        }
        close FILE;
    } elsif (m{^([0-9,]+)=(.*)}) {
        my ($what, $ifmt) = ($1, $2);
        print STDERR "warning: you have not given me a template (src=FILE)\n",
          "warning:   before reading race definitions; result may not be what\n",
          "warning:   you want.\n"
            if !@output_template && !$template_gripe;
        $template_gripe = 1;
        foreach my $race (split /,/, $what) {
            die "$race: must be number between 1 and 11\n"
              if ($race < 1 || $race > 11);
            my $name = sprintf($ifmt, $race);
            open FILE, "< $name"
              or die "$name: $!\n";
            while (<FILE>) {
                chomp;
                if (/^\s*$/ || /^\s*\#/) {
                    # nix
                } elsif (/^\s*(\w+)\s*=\s*(.*)/) {
                    my ($opt, $value) = ($1, $2);
                    $value =~ s/\#.*//;
                    $value =~ s/\s+$//;
                    set_option ($opt, $race, $value);
                } else {
                    die "syntax error: '$_'";
                }
            }
            close FILE;
        }
    } else {
        die "invalid parameter: '$_'\n";
    }
}

# Output

print STDERR "warning: you have not given me a template (src=FILE)\n",
  "warning:   outputing everything in alphabetical order\n"
  if !@output_template && !$template_gripe;
$template_gripe = 1;
foreach (@output_template) {
    if (m{^\+(.*)}) {
        print "$1\n";
        next;
    }

    my $value = $opt_value{$_};
    defined($value) or die "huh? no value for $_";
    printf "  %-35s = %s\n", $opt_pretty{$_}, $value;
    $opt_modified{$_} = 0;
}

my $head = 0;
foreach (sort keys %opt_modified) {
    if ($opt_modified{$_}) {
        print "\n% PHOST\n\n" unless $head;
        $head = 1;
        printf "  %-35s = %s\n", $opt_pretty{$_}, $opt_value{$_};
    }
}

sub set_option {
    my ($name, $race, $value) = @_;
    my $lname = lc($name);

    if (!exists $opt_is_race_array{$lname}) {
        # option is NOT an array indexed by race.
        if (exists $opt_modified{$lname} and !is_equivalent ($opt_value{$lname}, $value)) {
            print STDERR "warning: '$name' was set to '$opt_value{$lname}'\n",
              "warning:   but should later be set to '$value'\n";
        } else {
            if (!exists $opt_pretty{$lname}) {
                print STDERR "warning: there is no option '$name', inventing one\n";
                $opt_pretty{$lname} = $name;
            }
            $opt_value{$lname} = $value;
            $opt_modified{$lname} = 1;
        }
    } else {
        # modify an array in-place
        my @values = split /,\s*/, $opt_value{$lname};
        if (@values == 1 && is_equivalent($value, $values[0])) {
            # no need to modify it
        } else {
            while (@values < 11) { push @values, $values[-1] }
            $values[$race-1] = $value;
            $opt_value{$lname} = join ', ', @values;
            $opt_modified{$lname} = 1;
        }
    }
}

sub is_equivalent {
    return canonic($_[0]) eq canonic($_[1]);
}

sub canonic {
    my $v = lc(shift);
    $v =~ s/\s+//g;
    if ($v eq 'true') {
        'yes';
    } elsif ($v eq 'false') {
        'no';
    } else {
        $v;
    }
}
